#include <stdio.h> // for file I/O
#include <stdlib.h> // for exit()
#include <string.h> // for strncmp()
#include <memory.h> // for memset()
#pragma warning(disable : 4786) // don't warn about oversized STL debug names
#include <map> // STL template class
#include <vector> // STL template class
#if !defined (STLPORT) || defined(__STL_USE_NAMESPACES)
using namespace std; // for the SGI implementation of STL
#endif

#include "rvf_format.h" // my file format definitions

#ifndef INT_MAX
#define INT_MAX ((int)(~(unsigned)0 >> 1))
#endif
#ifndef INT_MIN
#define INT_MIN (-INT_MAX-1)
#endif

// constants
#define LCD_WIDTH 112
#define LCD_HEIGHT 64
#define RVF_FRAME_SIZE (LCD_WIDTH*LCD_HEIGHT/8)

// endian conversion macros
#define SWAP2(x) ((((unsigned)(x)>>8) & 0x00ff) | (((unsigned)(x)<<8) & 0xff00))
#define SWAP4(x) ((((unsigned)(x)>>24) & 0x000000ff) |\
                  (((unsigned)(x)>>8)  & 0x0000ff00) |\
                  (((unsigned)(x)<<8)  & 0x00ff0000) |\
                  (((unsigned)(x)<<24) & 0xff000000))

// 1:1 audio header from file
typedef struct
{	// note that bitfields are not necessarily portable
	// first byte
	unsigned char syncword_11_4 : 8;
	// second byte
	unsigned char protection_bit : 1;
	unsigned char layer : 2;
	unsigned char version : 2;
	unsigned char syncword_2_0 : 3;
	// third byte
	unsigned char private_bit : 1;
	unsigned char padding_bit : 1;
	unsigned char sampling_frequency : 2;
	unsigned char bitrate_index : 4;
	// fourth byte
	unsigned char emphasis : 2;
	unsigned char original_copy : 1;
	unsigned char copyright : 1;
	unsigned char mode_extension : 2;
	unsigned char mode : 2;
} tAudioHeader;

// decoded audio header info
typedef struct
{
	int bitrate; // in bits per second
	int samplerate; // in samples/second
	int framesize; // in bytes
	double frametime; // duration in seconds
	
	// context info:
	double abs_time; // absolute time of frame
	size_t filepos; // position in file
    size_t streampos; // position from audio start
} tHeaderInfo;


// command line configuration: what shall we do?
struct
{
	char* szVideoFile;
	char* szAudioFile;
    char* szOutFile;
    double dPlayFPS; // target speed
    double dVideoPreroll; // how much do you want video to be ahead
} gCfg;


// some statistics which are gathered while we work along
struct
{
    unsigned uBitrateAvg; // average audio bitrate
    unsigned uBitratePeak; // peak audio bitrate
    unsigned uAudioStart; // index of first audio block
    unsigned uVideoStart; // index of first video block
	unsigned uAudioPayloadSize; // total # of audio bytes
    int nMinAssoc; // minimum
    int nMaxAssoc; // and maximum of how far a video frame can be away
} gStat;

// map from audio file position to MPEG audio frame info
typedef map<size_t, tHeaderInfo, less<size_t>, allocator<tHeaderInfo> > tAudioMap;

// map from time to video frame position in file
typedef map<double, size_t, less<double>, allocator<size_t> > tVideoMap;


// implementation

// parse the command line (filenames and options)
int ProcessCmdLine(int argc, char* argv[])
{
	char* szValue;
    argc--; // exclude our name
	argv++;

	if (argc < 3)
	{
		printf("Usage: rvf_mux <videofile> <audiofile> <outfile> [-option {parameter}]\n");
		printf("  Multiplexes a Rockbox video file with a .mp3 (or .mp2/.mpa) audio file\n");
		printf("  into a new file containing both video and audio.\n");
		printf("  Video input format: YUV 4:2:0 planar (from avitoyuv.exe), 112*64\n");
		printf("Possible options are: (one letter for options is enough)\n");
		printf("-play_fps <fps>           FPS of Rockbox playback, default is 67.0\n");
		//printf("-video_preroll <seconds>  Video preroll, default is 0.0\n");
		printf("\n");
		printf("Example: \n");
		printf("rvf_mux doom.rvf doom.mp3 doom_av.rvf\n");
		printf("  joins 'doom.rvf' and 'doom.mp3' into a new file 'doom_av.rvf'\n");

		exit (0);
	}

    gCfg.szVideoFile = *argv++;
    argc--;

    gCfg.szAudioFile = *argv++;
    argc--;

    gCfg.szOutFile = *argv++;
    argc--;

	while (argc)
	{
		if (!strncmp("-play_fps", *argv, 2))
		{
			szValue = *++argv;
			if (--argc <= 0 || **argv == '-')
			{	
				printf("No argument given for option %s, aborting.\n", argv[-1]);
				exit (-2);
			}
            gCfg.dPlayFPS = atof(szValue);
		}
/*
        else if (!strncmp("-video_preroll", *argv, 2))
		{
			szValue = *++argv;
			if (--argc <= 0 || **argv == '-')
			{	
				printf("No argument given for option %s, aborting.\n", argv[-1]);
				exit (-2);
			}
            gCfg.dVideoPreroll = atof(szValue);
		}
*/
		else
		{
			printf("Unknown option %s, aborting. Use 'rvf_mux' without arguments for help.\n", *argv);
			exit(-1);
		}

		argv++;
		argc--;
	}
	
	return 0;
}


// helper function to read a certain fraction of a file
size_t ReadFrom(void *buffer, size_t size, size_t pos, FILE *stream )
{
	if (fseek(stream, pos, SEEK_SET) != 0)
		return 0;

	return fread(buffer, size, 1, stream);
}


// decode a raw audio header into a struct, return frame size or 0 if invalid
int DecodeAudioHeader(tAudioHeader* pAudioHeader, tHeaderInfo* pInfo)
{
	int nLayer;
	static const int aSamplerate[4][4] = 
	{
		{11025, 12000, 8000, 0},  // MPEG version 2.5
		{0, 0, 0, 0},             // illegal
		{22050, 24000, 16000, 0}, // MPEG version 2
		{44100, 48000, 32000, 0}  // MPEG version 1
	};
	static const int aBitRates[2][4][16] = 
	{
		{	// version 2 bitrate table
			{  0,	  0,     0,     0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      0, 0 },
			{  0, 32000, 48000, 56000,  64000,  80000,  96000, 112000, 128000, 144000, 160000, 176000, 192000, 224000, 256000, 0 },
			{  0,  8000, 16000, 24000,  32000,  40000,  48000,  56000,  64000,  80000,  96000, 112000, 128000, 144000, 160000, 0 },
			{  0,  8000, 16000, 24000,  32000,  40000,  48000,  56000,  64000,  80000,  96000, 112000, 128000, 144000, 160000, 0 }
		},
		{	// version 1 bitrate table
			{  0,     0,     0,     0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      0, 0 },
			{  0, 32000, 64000, 96000, 128000, 160000, 192000, 224000, 256000, 288000, 320000, 352000, 384000, 416000, 448000, 0 },
			{  0, 32000, 48000, 56000,  64000,  80000,  96000, 112000, 128000, 160000, 192000, 224000, 256000, 320000, 384000, 0 },
			{  0, 32000, 40000, 48000,  56000,  64000,  80000,  96000, 112000, 128000, 160000, 192000, 224000, 256000, 320000, 0 }
		},
	};
	
	if (pAudioHeader->syncword_11_4 != 0xFF || pAudioHeader->syncword_2_0 != 0x7)
		return 0;

	pInfo->samplerate = aSamplerate[pAudioHeader->version][pAudioHeader->sampling_frequency];
	if (pInfo->samplerate == 0) // invalid
		return 0;

	nLayer = 4 - pAudioHeader->layer; //  Get the layer so we can work out the bit rate
	pInfo->bitrate = aBitRates[pAudioHeader->version & 1][nLayer][pAudioHeader->bitrate_index];
	if (pInfo->bitrate == 0) // invalid
		return 0;

	if (nLayer == 1)
	{
		pInfo->framesize = (12 * pInfo->bitrate / pInfo->samplerate + pAudioHeader->padding_bit) * 4;
		pInfo->frametime = 384.0 / (double)pInfo->samplerate;
	}
	else 
	{
		pInfo->framesize = 144 * pInfo->bitrate / pInfo->samplerate;
		pInfo->frametime = 1152.0 / (double)pInfo->samplerate;
		if (pAudioHeader->version != 3) // not version 1:
		{
			pInfo->framesize /= 2; // half for version 2 and 2.5
			pInfo->frametime /= 2.0; // half the time for version 2 (LSF)
		}
		pInfo->framesize += pAudioHeader->padding_bit;
	}

	return pInfo->framesize;
}


// bitswap audio bytes, LSB becomes MSB and vice versa
int BitswapAudio (unsigned char* pDest, unsigned char* pSrc, size_t len)
{
    static const Lookup[256] = 
    {
        0x00,0x80,0x40,0xC0,0x20,0xA0,0x60,0xE0,0x10,0x90,0x50,0xD0,0x30,0xB0,0x70,0xF0,
        0x08,0x88,0x48,0xC8,0x28,0xA8,0x68,0xE8,0x18,0x98,0x58,0xD8,0x38,0xB8,0x78,0xF8,
        0x04,0x84,0x44,0xC4,0x24,0xA4,0x64,0xE4,0x14,0x94,0x54,0xD4,0x34,0xB4,0x74,0xF4,
        0x0C,0x8C,0x4C,0xCC,0x2C,0xAC,0x6C,0xEC,0x1C,0x9C,0x5C,0xDC,0x3C,0xBC,0x7C,0xFC,
        0x02,0x82,0x42,0xC2,0x22,0xA2,0x62,0xE2,0x12,0x92,0x52,0xD2,0x32,0xB2,0x72,0xF2,
        0x0A,0x8A,0x4A,0xCA,0x2A,0xAA,0x6A,0xEA,0x1A,0x9A,0x5A,0xDA,0x3A,0xBA,0x7A,0xFA,
        0x06,0x86,0x46,0xC6,0x26,0xA6,0x66,0xE6,0x16,0x96,0x56,0xD6,0x36,0xB6,0x76,0xF6,
        0x0E,0x8E,0x4E,0xCE,0x2E,0xAE,0x6E,0xEE,0x1E,0x9E,0x5E,0xDE,0x3E,0xBE,0x7E,0xFE,
        0x01,0x81,0x41,0xC1,0x21,0xA1,0x61,0xE1,0x11,0x91,0x51,0xD1,0x31,0xB1,0x71,0xF1,
        0x09,0x89,0x49,0xC9,0x29,0xA9,0x69,0xE9,0x19,0x99,0x59,0xD9,0x39,0xB9,0x79,0xF9,
        0x05,0x85,0x45,0xC5,0x25,0xA5,0x65,0xE5,0x15,0x95,0x55,0xD5,0x35,0xB5,0x75,0xF5,
        0x0D,0x8D,0x4D,0xCD,0x2D,0xAD,0x6D,0xED,0x1D,0x9D,0x5D,0xDD,0x3D,0xBD,0x7D,0xFD,
        0x03,0x83,0x43,0xC3,0x23,0xA3,0x63,0xE3,0x13,0x93,0x53,0xD3,0x33,0xB3,0x73,0xF3,
        0x0B,0x8B,0x4B,0xCB,0x2B,0xAB,0x6B,0xEB,0x1B,0x9B,0x5B,0xDB,0x3B,0xBB,0x7B,0xFB,
        0x07,0x87,0x47,0xC7,0x27,0xA7,0x67,0xE7,0x17,0x97,0x57,0xD7,0x37,0xB7,0x77,0xF7,
        0x0F,0x8F,0x4F,0xCF,0x2F,0xAF,0x6F,0xEF,0x1F,0x9F,0x5F,0xDF,0x3F,0xBF,0x7F,0xFF,
    };

    while (len--)
        *pDest++ = Lookup[*pSrc++];

    return 0;
}


// map the mp3 file, associate file position to MPEG header info
double MapAudio(FILE* pInFile, tAudioMap* pMap)
{
	size_t filesize;
	size_t filepos = 0;
	size_t filestart; // offset of first audio frame
	double time = 0.0;
    double bitrate_avg = 0;

	tAudioHeader AudioHeader;
	tHeaderInfo HeaderInfo;
	bool bSynced = false;

#ifdef _DEBUG
    printf("filepos, framesize, time, bitrate, samplerate\n");
#endif

    fseek(pInFile, 0, SEEK_END);
	filesize = ftell(pInFile);
	fseek(pInFile, 0, SEEK_SET);

	while (filepos < filesize - sizeof(AudioHeader))
	{
		ReadFrom(&AudioHeader, sizeof(AudioHeader), filepos, pInFile);
		if (AudioHeader.syncword_11_4 == 0xFF 
			&& AudioHeader.syncword_2_0 == 0x7
			&& DecodeAudioHeader(&AudioHeader, &HeaderInfo))
		{
			if (!bSynced && filepos < filesize - HeaderInfo.framesize - sizeof(AudioHeader))
			{
				// check next header
				tAudioHeader NextAudioHeader;
				tHeaderInfo NextHeaderInfo;
				ReadFrom(&NextAudioHeader, sizeof(NextAudioHeader), filepos + HeaderInfo.framesize, pInFile);
				if (NextAudioHeader.syncword_11_4 == 0xFF 
					&& NextAudioHeader.syncword_2_0 == 0x7
					&& DecodeAudioHeader(&NextAudioHeader, &NextHeaderInfo))
				{
					// check for extra VBR header frame
					char HeaderVBR[4];
					const int vbrpos[2][2] = {{13, 21}, {21, 36}};
					ReadFrom(&HeaderVBR, sizeof(HeaderVBR), filepos + vbrpos[AudioHeader.version&1][AudioHeader.mode != 3], pInFile);
					if (strncmp(HeaderVBR, "Xing", sizeof(HeaderVBR)) == 0
					 || strncmp(HeaderVBR, "VBRI", sizeof(HeaderVBR)) == 0
					 || strncmp(HeaderVBR, "Info", sizeof(HeaderVBR)) == 0)
					{	// skip this frame
						filepos += HeaderInfo.framesize;
					}
					else
					{
						bSynced = true;	
						if (pMap->empty()) // not set yet?
							filestart = filepos;
						else
							printf("Audio sync error in file %s at %d, possibly corrupted.\n", gCfg.szAudioFile,  filepos);
					}
				}
				else // false alert
					filepos++; // check from next byte on
			}

			if (bSynced)
			{
				// add the context info
				HeaderInfo.filepos = filepos;
				HeaderInfo.streampos = filepos - filestart;
				HeaderInfo.abs_time = time;
				pMap->insert(tAudioMap::value_type(filepos, HeaderInfo)).first; // add to map
#ifdef _DEBUG
                //printf("filepos, framesize, time, bitrate, samplerate\n"
                printf("0x%X, %d, %f, %d, %d\n",  
                    HeaderInfo.filepos,
                    HeaderInfo.framesize,
                    HeaderInfo.abs_time,
                    HeaderInfo.bitrate,
                    HeaderInfo.samplerate);
#endif
                // statistics
				gStat.uAudioPayloadSize += HeaderInfo.framesize;
                bitrate_avg += HeaderInfo.bitrate; // need to divide later
                if(HeaderInfo.bitrate > gStat.uBitratePeak)
                    gStat.uBitratePeak = HeaderInfo.bitrate;

				filepos += HeaderInfo.framesize;
				time += HeaderInfo.frametime;
			}
		}
		else
		{
			bSynced = false;
			filepos++; // check from next byte on
		}
	}

    // update + return statistics
    bitrate_avg /= pMap->size(); // divide the sum by the number of entries
    gStat.uBitrateAvg = (unsigned long)(bitrate_avg + 0.5); // round it

	return time;
}


// map the input rvf file, associate time to file positions
int MapVideo(FILE* pInFile, tVideoMap* pMap)
{
	size_t filesize;
	fseek(pInFile, 0, SEEK_END);
	filesize = ftell(pInFile);
	fseek(pInFile, 0, SEEK_SET);
	size_t filepos = 0;
	double dTime = 0.0;

	tFileHeader FileHeader;
	ReadFrom(&FileHeader, sizeof(FileHeader), 0, pInFile);
	if (FileHeader.magic == SWAP4(HEADER_MAGIC)
		&& FileHeader.audio_format != SWAP4(AUDIOFORMAT_NO_AUDIO))
	{	// mapping for a muxed file, requires parsing it
		tAudioFrameHeader AudioFrameHeader;
//#define CHECK_MP3
#ifdef CHECK_MP3 // test code
		FILE* pDump = fopen("dump.mp3", "wb");
#endif // end of test code
		
		filepos = RVF_FRAME_SIZE; // don't process the file header frame
		int frame = 0; // video frame counter
		while (filepos < filesize)
		{
			ReadFrom(&AudioFrameHeader, sizeof(AudioFrameHeader), filepos, pInFile);
			if (AudioFrameHeader.magic == SWAP4(AUDIO_MAGIC)) // audio frame?
			{	
#ifdef CHECK_MP3 // test code
				unsigned char buffer[RVF_FRAME_SIZE - sizeof(AudioFrameHeader)];
				ReadFrom(buffer, sizeof(buffer), filepos + sizeof(AudioFrameHeader), pInFile);
				BitswapAudio(buffer, buffer, sizeof(buffer));
				fwrite(buffer, sizeof(buffer), 1, pDump);
				if (buffer[SWAP2(AudioFrameHeader.frame_start)] != 0xFF)
					printf("no mp3 frame start at %d\n", filepos + sizeof(AudioFrameHeader) + SWAP2(AudioFrameHeader.frame_start));
				if (buffer[SWAP2(AudioFrameHeader.frame_end)] != 0xFF
					&& SWAP2(AudioFrameHeader.frame_end) != sizeof(buffer))
					printf("no mp3 frame end at %d\n", filepos + sizeof(AudioFrameHeader) + SWAP2(AudioFrameHeader.frame_end));
#endif // end of test code
				
				filepos += RVF_FRAME_SIZE; // behind that one
				
				// an audio frame points to the next one, 
				// so we can mark all intermediate frames as video without touching them
				for (int i=1; i<AudioFrameHeader.next_block; i++)
				{
					dTime = (double)frame++ / gCfg.dPlayFPS; 
					pMap->insert(tVideoMap::value_type(dTime, filepos)).first; // add to map
					filepos += RVF_FRAME_SIZE;
				}
			}
			else // was a video frame
			{
				dTime = (double)frame++ / gCfg.dPlayFPS; 
				pMap->insert(tVideoMap::value_type(dTime, filepos)).first; // add to map
				filepos += RVF_FRAME_SIZE;
			}
		} // while (filepos < filesize)

#ifdef CHECK_MP3 // test code
		fclose(pDump);
#endif // end of test code

	} // if muxed
	else // simple if the rvf file is not yet multiplexed.
	{
		int frames = filesize/RVF_FRAME_SIZE;
		for (int i = 0; i < frames; i++)
		{
			dTime = (double)i / gCfg.dPlayFPS; 
			pMap->insert(tVideoMap::value_type(dTime, filepos)).first; // add to map
			filepos += RVF_FRAME_SIZE;
		}
	}
	return pMap->size(); // return # of frames
}


// create the destination file from timing info
int MultiplexAudioVideo(FILE* pAudioFile, tAudioMap* pAudioMap, 
						FILE* pVideoFile, tVideoMap* pVideoMap,
						FILE* pOutFile)
{
    double dAudioTime, dVideoTime;
	dAudioTime = 0.0;
	dVideoTime = 0.0;
    bool bFirstAudio = true;
    bool bFirstVideo = true;

    vector<tHeaderInfo> DestinationAudio; // list for linking audio blocks

    tAudioMap::iterator ita = pAudioMap->begin();
	tVideoMap::iterator itv = pVideoMap->begin();

    fseek(pAudioFile, ita->second.filepos, SEEK_SET); // go to first frame

    // prepare buffer
    unsigned char buffer[RVF_FRAME_SIZE];
    size_t audio_filepos = ftell(pAudioFile); // at first frame
    size_t out_filepos = ftell(pOutFile); // past the file header

    // prepare audio header
	tAudioFrameHeader AudioFrameHeader;
    memset(&AudioFrameHeader, 0, sizeof(AudioFrameHeader));
    AudioFrameHeader.magic = SWAP4(AUDIO_MAGIC);
    
	while (ita != pAudioMap->end() || itv != pVideoMap->end())
	{
        while (ita != pAudioMap->end() 
            && (dAudioTime + gCfg.dVideoPreroll <= dVideoTime // as long as due
            || itv == pVideoMap->end())) // or all the rest if video file exhausted
		{
		    if (bFirstAudio)
            {   // return the position
                bFirstAudio = false;
                gStat.uAudioStart = out_filepos;
            }
            
            // lookup the time for this audio block
            dAudioTime = pAudioMap->lower_bound(audio_filepos)->second.abs_time;

            // add audio block to list for later linking
            DestinationAudio.push_back(ita->second);
            DestinationAudio.back().filepos = out_filepos; // new pos

            // read payload
            audio_filepos += fread(buffer, 1, RVF_FRAME_SIZE - sizeof(AudioFrameHeader), pAudioFile);

			// determine end of last frame in buffer
            ita = pAudioMap->upper_bound(audio_filepos); // beyound current position
			ita--; // this gives the last one starting in this block or exactly with the next
    		// FixMe: this modulo only works if a block always contains a full audio frame
			int frame_end = ita->second.streampos % (RVF_FRAME_SIZE - sizeof(tAudioFrameHeader));
			if (frame_end == 0)
				frame_end = RVF_FRAME_SIZE - sizeof(tAudioFrameHeader);
			AudioFrameHeader.frame_end = SWAP2(frame_end);
            
            // write header, to be updated later
            out_filepos += fwrite(&AudioFrameHeader, 1, sizeof(AudioFrameHeader), pOutFile);

            // convert to our native format
            BitswapAudio(buffer, buffer, RVF_FRAME_SIZE - sizeof(tAudioFrameHeader));

            // write payload into block
            out_filepos += fwrite(buffer, 1, RVF_FRAME_SIZE - sizeof(tAudioFrameHeader), pOutFile);

            // lookup for the next block
            ita = pAudioMap->lower_bound(audio_filepos);
		}
		
        if (itv != pVideoMap->end())
        {
		    if (bFirstVideo)
            {   // return the position
                bFirstVideo = false;
                gStat.uVideoStart = out_filepos;
            }

            // read video frame
            ReadFrom(buffer, RVF_FRAME_SIZE, itv->second, pVideoFile);

            itv->second = out_filepos; // change into destination position
		    
            // write video frame
            out_filepos += fwrite(buffer, 1, RVF_FRAME_SIZE, pOutFile);
		    
            dVideoTime += 1.0 / gCfg.dPlayFPS;
            itv++;
        }
	}

    // link the audio blocks
    printf("Linking audio blocks\n");
    gStat.nMinAssoc = INT_MAX; // set statistics to extreme values
    gStat.nMaxAssoc = INT_MIN; // so they will get updated
    for (int i=0; i<DestinationAudio.size(); i++)
    {
        long diff;

        // read the audio header, to re-use what's been done before
		ReadFrom(&AudioFrameHeader, sizeof(AudioFrameHeader), DestinationAudio[i].filepos, pOutFile);

		if (i != 0) // not the first
        {	// distance to the previous
            diff = DestinationAudio[i].filepos - DestinationAudio[i-1].filepos;
            AudioFrameHeader.previous_block = (unsigned char)(diff / RVF_FRAME_SIZE);
        }

        if (i < DestinationAudio.size() - 1) // not the last
        {	// distance to the next
            diff = DestinationAudio[i+1].filepos - DestinationAudio[i].filepos;
            AudioFrameHeader.next_block = (unsigned char)(diff / RVF_FRAME_SIZE);
        }

		// link to video frame with closest timestamp
        itv = pVideoMap->lower_bound(DestinationAudio[i].abs_time);
        if (itv == pVideoMap->end())
        {
            diff = 0; // none to be associated
        }
        else
        {
		    diff = itv->second - DestinationAudio[i].filepos;
            if (diff < gStat.nMinAssoc) // remember minimum and maximum
                gStat.nMinAssoc = diff;
            if (diff > gStat.nMaxAssoc)
                gStat.nMaxAssoc = diff;
        }
        AudioFrameHeader.associated_video = SWAP2((short)(diff/RVF_FRAME_SIZE));

        // offset of frame start
		// FixMe: this modulo only works if a block always contains a full audio frame
		AudioFrameHeader.frame_start = 
			SWAP2(DestinationAudio[i].streampos % (RVF_FRAME_SIZE - sizeof(tAudioFrameHeader)));

        // write the updated audio header
		fseek(pOutFile, DestinationAudio[i].filepos, SEEK_SET);
        fwrite(&AudioFrameHeader, sizeof(AudioFrameHeader), 1, pOutFile);
    }

	return 0;
}


int main (int argc, char** argv)
{
    // set defaults
    memset(&gStat, 0, sizeof(gStat));
    memset(&gCfg, 0, sizeof(gCfg));
    gCfg.dPlayFPS = 67.0;
    gCfg.dVideoPreroll = 0.0;

    ProcessCmdLine(argc, argv); // what to do

    // round like the hardware has to, with prescaler = 4
    int frameclocks = (int)((double)CLOCK/gCfg.dPlayFPS) / 4;
    gCfg.dPlayFPS = (double)CLOCK / (double)frameclocks / 4.0;

	FILE* pVideoFile;
    pVideoFile = fopen(gCfg.szVideoFile, "rb");
    if (pVideoFile == NULL)
    {
        printf("cannot open video input file %s\n", gCfg.szVideoFile);
        exit(2);
    }

    FILE* pAudioFile;
    pAudioFile = fopen(gCfg.szAudioFile, "rb");
    if (pAudioFile == NULL)
    {
        printf("cannot open audio input file %s\n", gCfg.szAudioFile);
        exit(2);
    }

    FILE* pOutFile;
    pOutFile = fopen(gCfg.szOutFile, "w+b");
    if (pOutFile == NULL)
    {
        printf("cannot open output file %s, gCfg.szOutFile)\n");
        exit(3);
    }

	// map the audio frames
    tAudioMap AudioMap;
    double dAudioTime;
    printf("Creating map of audio file %s\n", gCfg.szAudioFile);
    dAudioTime = MapAudio(pAudioFile, &AudioMap);

/*
	// debug dump
	tAudioMap::iterator ita = AudioMap.begin();
	while(ita != AudioMap.end()) 
	{
		size_t filepos = ita->first;
		tHeaderInfo HeaderInfo = ita->second;
		ita++;

		printf("%08X: %d Hz, %6d bps, %d bytes, %.2f sec\n", 
			filepos, HeaderInfo.samplerate, HeaderInfo.bitrate, HeaderInfo.framesize, HeaderInfo.abs_time);
	}
*/

	tVideoMap VideoMap;
    int nVideoFrames;
    printf("Creating map of video file %s\n", gCfg.szVideoFile);
	nVideoFrames = MapVideo(pVideoFile, &VideoMap); // map the video frames
/*
	// debug dump
	tVideoMap::iterator itv = VideoMap.begin();
	while(itv != VideoMap.end()) 
	{
		double time = itv->first;
		size_t filepos = itv->second;
		itv++;

		printf("%.2f sec: %8d\n", time, filepos);
	}
*/
	
	
	// finally, mux it together
    printf("Multiplexing into output file %s\n", gCfg.szOutFile);
	fseek(pOutFile, RVF_FRAME_SIZE, SEEK_SET); // befind the later header block
    MultiplexAudioVideo(pAudioFile, &AudioMap, pVideoFile, &VideoMap, pOutFile);


	// update + write the file header
    unsigned long uVideoBitrate = (unsigned long)(0.5 + LCD_WIDTH * LCD_HEIGHT * gCfg.dPlayFPS);
	fseek(pOutFile, 0, SEEK_SET); // start of file again
	tFileHeader FileHeader;
    memset(&FileHeader, 0, sizeof(FileHeader));
	FileHeader.magic     = SWAP4(HEADER_MAGIC);
	FileHeader.version   = SWAP4(FILEVERSION);
	FileHeader.flags     = SWAP4(0);
	FileHeader.blocksize = SWAP4(RVF_FRAME_SIZE);
    FileHeader.bps_average = SWAP4(uVideoBitrate + gStat.uBitrateAvg);
    FileHeader.bps_peak    = SWAP4(uVideoBitrate + gStat.uBitratePeak);

	FileHeader.video_format    = SWAP4(VIDEOFORMAT_RAW);
    FileHeader.video_1st_frame = SWAP4(gStat.uVideoStart);
    FileHeader.video_duration  = SWAP4(1000.0 * nVideoFrames / gCfg.dPlayFPS + 0.5);
	FileHeader.video_payload_size = SWAP4(nVideoFrames * RVF_FRAME_SIZE);
    FileHeader.video_bitrate   = SWAP4(uVideoBitrate);
	FileHeader.video_frametime = SWAP4((double)CLOCK/gCfg.dPlayFPS);
    FileHeader.video_preroll   = SWAP4((long)((double)CLOCK*gCfg.dVideoPreroll));
	FileHeader.video_width     = SWAP4(LCD_WIDTH);
	FileHeader.video_height    = SWAP4(LCD_HEIGHT);
	
    FileHeader.audio_format         = SWAP4(AUDIOFORMAT_MP3_BITSWAPPED);
    FileHeader.audio_1st_frame      = SWAP4(gStat.uAudioStart);
    FileHeader.audio_duration       = SWAP4(dAudioTime * 1000 + 0.5);
	FileHeader.audio_payload_size   = SWAP4(gStat.uAudioPayloadSize);
    FileHeader.audio_avg_bitrate    = SWAP4(gStat.uBitrateAvg);
    FileHeader.audio_peak_bitrate   = SWAP4(gStat.uBitratePeak);
	FileHeader.audio_headersize     = SWAP4(sizeof(tAudioFrameHeader));
    FileHeader.audio_min_associated = SWAP4(gStat.nMinAssoc);
    FileHeader.audio_max_associated = SWAP4(gStat.nMaxAssoc);

    fwrite(&FileHeader, sizeof(FileHeader), 1, pOutFile); // write it out

    
    fclose(pVideoFile);
    fclose(pAudioFile);
    fclose(pOutFile);

    return 0;
}
